@@ -1,5 +1,7 @@
 from .version import version as __version__
 from .core import predict, get_output_path, process_file, load_classifier, \
                   format_pred, format_pred_batch, compute_pcen, \
+                  load_taxonomy, \
+                  get_best_candidates, get_batch_best_candidates, \
                   get_taxonomy_node, get_taxonomy_path, get_model_path, \
                   get_pcen_settings, batch_generator, DEFAULT_MODEL_NAME
